import jax
import jax.numpy as jnp
import haiku as hk



class ActivityRegularization(hk.Module):
    """
    Add state to the SNN to track the average number of spikes emitted per neuron per batch.

    Adding this to a network requires using the Haiku transform_with_state transform, which will also return an initial regularization state vector.
    This blank initial vector can be reused and is provided as the second arg to the SNN's apply function. 
    """

    def __init__(self, name="ActReg"):
        super().__init__(name=name)
        
    def __call__(self, spikes):
        spike_count = hk.get_state("spike_count", spikes.shape, init=jnp.zeros, dtype=spikes.dtype)
        hk.set_state("spike_count", spike_count + spikes) 
        return spikes



def tanh(k=1):
    """
        Hyperbolic Tangent activation. Very simple.

    """
    def g(x):
        kx = k * x
        return 4 / (jnp.exp(-kx) + jnp.exp(kx))**2
    return jax.jit(g)




def boxcar(width=2, height=0.5):
    """
        Boxcar activation. Very simple.

    """
    k = width / 2
    h = height
    def g(x):
        return h * jnp.heaviside(-(jnp.abs(x) - k), 0)
    return jax.jit(g)




def triangular(k=0.5):
    """
        Triangular activation inspired by Esser et. al. Very simple. https://arxiv.org/abs/1603.08270

    """
    def g(x):
        return jnp.maximum(0, 1-jnp.abs(k*x))
    return jax.jit(g)





def arctan(k=2):
    """
    This class implements the Arctangent surrogate gradient activation function for a spiking neuron.
    
    The Arctangent function is a smooth function that approximates the step function. 
    It is used as a surrogate gradient for the step function in the context of spiking neurons. 
    The surrogate gradient is used during the backpropagation process to update the weights of the neuron.
    
    The Arctangent function returns a value between -pi/2 and pi/2 for inputs in the range of -Infinity to Infinity.
    It is often used in the context of spiking neurons because it provides a smooth approximation to the step function 
    that is differentiable everywhere, which is a requirement for gradient-based optimization methods.
    
    Attributes:
        scale_factor: A scaling factor that can be used to adjust the steepness of the 
                      Arctangent function. Default is 2.
    """
    def g(U):
        x = jnp.pi * U * k
        return (1 / (1+x**2)) / jnp.pi

    return jax.jit(g)




def sigmoid(k=4):
    """
    This class implements the Sigmoid surrogate gradient activation function for a spiking neuron.
    
    The Sigmoid function is a smooth function that approximates the step function. 
    It is used as a surrogate gradient for the step function in the context of spiking neurons. 
    The surrogate gradient is used during the backpropagation process to update the weights of the neuron.
    
    The Sigmoid function returns a value between 0 and 1 for inputs in the range of -Infinity to Infinity.
    It is often used in the context of spiking neurons because it provides a smooth approximation to the step function 
    that is differentiable everywhere, which is a requirement for gradient-based optimization methods.
    
    Attributes:
        scale_factor: A scaling factor that can be used to adjust the steepness of the 
                      Sigmoid function. Default is 4.
    """
    sig = jax.grad(jax.nn.sigmoid)
    def g(x):
        v = sig # development found a really weird numerical instability with computing the derivative outright
        for axis in x.shape: # where the model wouldn't learn, but using jax.grad of the sigmoid function
            v = jax.vmap(v) # would learn despite the differences being < 1e-7 without a bias +/1.
        return v(4*x) # to make the grad transformed func work on any shape input, we have to use this ugly
    return jax.jit(g) # hack which vmaps to accomodate the shape. Luckily, it appears to not impact the compiled func.



def superspike(k=25):
    """
    This function implements the SuperSpike surrogate gradient activation function for a spiking neuron.
    
    The SuperSpike function is a smooth function that approximates the step function. 
    It is used as a surrogate gradient for the step function in the context of spiking neurons. 
    The surrogate gradient is used during the backpropagation process to update the weights of the neuron.
    
    The SuperSpike function is defined as 1/(1+k|U|)^2, where U is the input to the function and k is a scaling factor.
    It returns a value between 0 and 1 for inputs in the range of -Infinity to Infinity.
    
    It is often used in the context of spiking neurons because it provides a smooth approximation to the step function 
    that is differentiable everywhere, which is a requirement for gradient-based optimization methods.

    It is a fast approximation of the Sigmoid function adapted from:

    F. Zenke, S. Ganguli (2018) SuperSpike: Supervised Learning in Multilayer Spiking Neural Networks. Neural Computation, pp. 1514-1541.
    
    Attributes:
        scale_factor: A scaling factor that can be used to adjust the steepness of the 
                      SuperSpike function. Default is 25.
    """

    def g(x):
        return 1 / (1 + k*jnp.abs(x))**2
    return jax.jit(g)


# This could also just be changed to be a function which yields the proper VJP func...
def Axon(bwd=jax.jit(lambda x: 1), 
         fwd=jax.jit(lambda x: jnp.heaviside(x,0))):
    """
    This function serves as the activation function for the SNNs, allowing for custom definitions of both surrogate gradients for backwards
    passes as well as substitution of the Heaviside function for relaxations such as sigmoids. 

    In short, this function takes functions that define the backward and forward passes for the spiking activation function and fuses them
    into a single function using VJP, allowing for surrogate gradient methods as well as for using alternative forward activations when performing
    neuroevolution or ANN to SNN conversion. The user can also define and JIT their own function to use as a surrogate, or
    even use more dynamic functions that change over the course of training.

    The default behavior is a Heaviside forward activation with a stragiht through estimator surrogate gradient.

    Attributes:
        bwd: Function that calculates the gradient to be used in the backwards pass.
        fwd: Function that returns a value between 0 and 1. Default is Heaviside.
     
    """
        
    @jax.custom_vjp
    def f(U): # primal function
        return fwd(U)
        
    # returns value, grad context
    def f_fwd(U):
        return f(U), U
            
    # accepts context, primal val
    def f_bwd(U, grad):
        return (grad * bwd(U) , )
            
    f.defvjp(f_fwd, f_bwd)
    
    return jax.jit(f)


import tonic
from tonic import datasets, transforms
import torchvision as tv
from torch.utils.data import DataLoader, Subset
from sklearn.model_selection import train_test_split
from collections import namedtuple

import numpy as np
import jax
import jax.numpy as jnp


State = namedtuple("State", "obs labels")

# should add a class that takes a numpy dataset and reshapes it and compresses it...

# This should be changed to a higher-order function
class shift_augment:
    """
        Shift data augmentation tool. Rolls data along specified axes randomly up to a certain amount.
    """

    def __init__(self, max_shift=10, axes=(-1,)):
        self.max_shift = max_shift
        self.axes = axes

        @jax.jit
        def _shift(data, rng):
            shift = jax.random.randint(rng, (len(self.axes),), -self.max_shift, self.max_shift)
            return jnp.roll(data, shift, self.axes)

        self._shift = _shift


    def __call__(self, data, rng):
        return self._shift(data, rng)





# Here we scale the max probability to .75 so that we don't have inputs that are continually spiking.
def rate_code(data, steps, key, max_r=0.75):
    """Unrolls input data along axis 1 and converts to rate encoded spikes."""

    data = jnp.array(data, dtype=jnp.float16)
    unrolled_data = jnp.repeat(data, steps, axis=1)
    return jax.random.bernoulli(key, unrolled_data*max_r).astype(jnp.uint8)


class MNIST_loader(): # change this so that it just returns either rate or temporal mnist...
    """
    Dataloader for the MNIST dataset. The data is returned in a  non-temporal format so instead of applying
    jnp.unpack_bits on the data you need to apply the spyx.data.rate_code function to generate the SNN input.

    
    """

    # Change this to allow a config dictionary of 
    def __init__(self, time_steps=64, max_rate = 0.75, batch_size=32, val_size=0.3, subsample_data=1, key=jax.random.PRNGKey(0), download_dir='./MNIST'):
        key1, key2, key3 = jax.random.split(key, 3)
        self.sample_T = time_steps
        self.max_rate = max_rate
        self.val_size = val_size
        self.batch_size = batch_size
        self.obs_shape = (28,28)
        self.act_shape = tuple([10,])
        
        transform = tv.transforms.Compose([
            tv.transforms.Resize(self.obs_shape),
            tv.transforms.Grayscale(),
            tv.transforms.ToTensor(),
            tv.transforms.Normalize((0,), (1,)),
            lambda x: np.expand_dims(x, axis=-1)
            ])
        
        # fix this
        train_val_dataset = tv.datasets.MNIST("./data", train=True, download=True, transform=transform)
        test_dataset = tv.datasets.MNIST("./data", train=False, download=True, transform=transform)
        # create train/validation split here...
        # generate indices: instead of the actual data we pass in integers instead
        train_indices, val_indices = train_test_split(
        range(len(train_val_dataset)),
        test_size=self.val_size,
        random_state=0,
        shuffle=True
        )

        # to help with trying to do neuroevolution since the full dataset is a bit much for evolving convnets...
        train_indices = train_indices[:int(len(train_indices)*subsample_data)]
        val_indicies  = val_indices[:int(len(val_indices)*subsample_data)]
    
    
        train_split = Subset(train_val_dataset, train_indices)
        self.train_len = len(train_indices)

        train_dl = iter(DataLoader(train_split, batch_size=self.train_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=False))
        
        x_train, y_train = next(train_dl)
        self.x_train = jnp.packbits(rate_code(jnp.array(x_train, dtype=jnp.uint8), self.sample_T, key1), axis=1)
        self.y_train = jnp.array(y_train, dtype=jnp.uint8)
        ############################
        
        val_split = Subset(train_val_dataset, val_indices)
        self.val_len = len(val_indices)

        val_dl = iter(DataLoader(val_split, batch_size=self.val_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=False))
        
        x_val, y_val = next(val_dl)
        self.x_val = jnp.packbits(rate_code(jnp.array(x_val, dtype=jnp.uint8), self.sample_T, key2), axis=1)
        self.y_val = jnp.array(y_val, dtype=jnp.uint8)
        ##########################
        # Test set setup
        ##########################
        self.test_len = len(test_dataset)
        test_dl = iter(DataLoader(test_dataset, batch_size=self.test_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=True))
                
        x_test, y_test = next(test_dl)
        self.x_test = jnp.packbits(rate_code(jnp.array(x_test, dtype=jnp.uint8), self.sample_T, key3), axis=1)
        self.y_test = jnp.array(y_test, dtype=jnp.uint8)


        
        @jax.jit
        def _train_epoch(shuffle_key):
            cutoff = self.train_len % self.batch_size
            
            obs = jax.random.permutation(shuffle_key, self.x_train, axis=0)[:-cutoff] # self.x_train[:-cutoff]
            labels = jax.random.permutation(shuffle_key, self.y_train, axis=0)[:-cutoff] # self.y_train[:-cutoff]
            
            obs = jnp.reshape(obs, (-1, self.batch_size) + obs.shape[1:])
            labels = jnp.reshape(labels, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
            
        self.train_epoch = _train_epoch
            
        @jax.jit
        def _val_epoch():
            cutoff = self.val_len % self.batch_size
            
            x_val = self.x_val[:-cutoff]
            y_val = self.y_val[:-cutoff]
            
            obs = jnp.reshape(x_val, (-1, self.batch_size) + x_val.shape[1:])
            labels = jnp.reshape(y_val, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
        
        self.val_epoch = _val_epoch
        
        
        @jax.jit
        def _test_epoch():
            cutoff = self.test_len % self.batch_size
            
            x_test = self.x_test[:-cutoff]
            y_test = self.y_test[:-cutoff]
            
            obs = jnp.reshape(x_test, (-1, self.batch_size) + x_test.shape[1:])
            labels = jnp.reshape(y_test, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
        
        self.test_epoch = _test_epoch


###############################################

class NMNIST_loader():
    """
    Dataloading wrapper for the Neuromorphic MNIST dataset.
    """

    # Change this to allow a config dictionary of 
    def __init__(self, batch_size=32, sample_T = 40, data_frac = .33, val_size=0.3, download_dir='./NMNIST'):
           
        self.val_size = val_size
        self.batch_size = batch_size
        self.obs_shape = datasets.NMNIST.sensor_size
        self.act_shape = tuple([10,])
        
        transform =  transforms.Compose([
            transforms.ToFrame(sensor_size=self.obs_shape, 
                                       n_time_bins=sample_T),
            lambda x: np.packbits(x, axis=0)

        ])
        

        train_val_dataset = datasets.NMNIST("./data", first_saccade_only=True, train=True, transform=transform)
        test_dataset = datasets.NMNIST("./data", first_saccade_only=True, train=False, transform=transform)
        
        
        train_indices, val_indices = train_test_split(
        range(len(train_val_dataset)),
        test_size=self.val_size,
        random_state=314159,
        shuffle=True
        )
    
        train_indices = train_indices[:int(len(train_indices)*data_frac)]
        val_indices = val_indices[:int(len(val_indices)*data_frac)]
    
        train_split = Subset(train_val_dataset, train_indices)
        self.train_len = len(train_indices)

        train_dl = iter(DataLoader(train_split, batch_size=self.train_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=False))
        
        x_train, y_train = next(train_dl)
        self.x_train = jnp.array(x_train, dtype=jnp.uint8)
        self.y_train = jnp.array(y_train, dtype=jnp.uint8)
        ############################
        
        val_split = Subset(train_val_dataset, val_indices)
        self.val_len = len(val_indices)

        val_dl = iter(DataLoader(val_split, batch_size=self.val_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=False))
        
        x_val, y_val = next(val_dl)
        self.x_val = jnp.array(x_val, dtype=jnp.uint8)
        self.y_val = jnp.array(y_val, dtype=jnp.uint8)
        ##########################
        # Test set setup
        ##########################
        self.test_len = len(test_dataset)
        test_dl = iter(DataLoader(test_dataset, batch_size=self.test_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=True))
                
        x_test, y_test = next(test_dl)
        self.x_test = jnp.array(x_test, dtype=jnp.uint8)
        self.y_test = jnp.array(y_test, dtype=jnp.uint8)


        
        @jax.jit
        def _train_epoch(shuffle_key):
            cutoff = self.train_len % self.batch_size
            
            obs = jax.random.permutation(shuffle_key, self.x_train, axis=0)[:-cutoff] # self.x_train[:-cutoff]
            labels = jax.random.permutation(shuffle_key, self.y_train, axis=0)[:-cutoff] # self.y_train[:-cutoff]
            
            obs = jnp.reshape(obs, (-1, self.batch_size) + obs.shape[1:])
            labels = jnp.reshape(labels, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
            
        self.train_epoch = _train_epoch
            
        @jax.jit
        def _val_epoch():
            cutoff = self.val_len % self.batch_size
            
            x_val = self.x_val[:-cutoff]
            y_val = self.y_val[:-cutoff]
            
            obs = jnp.reshape(x_val, (-1, self.batch_size) + x_val.shape[1:])
            labels = jnp.reshape(y_val, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
        
        self.val_epoch = _val_epoch
        
        
        @jax.jit
        def _test_epoch():
            cutoff = self.test_len % self.batch_size
            
            x_test = self.x_test[:-cutoff]
            y_test = self.y_test[:-cutoff]
            
            obs = jnp.reshape(x_test, (-1, self.batch_size) + x_test.shape[1:])
            labels = jnp.reshape(y_test, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
        
        self.test_epoch = _test_epoch
        
###########################################################

# Builds 2D tensors from data, with the time axis being packed to save memory. 
class SHD2Raster():
    """ Tool for rastering SHD samples into frames. Packs bits along the temporal axis for memory efficiency. This means
        that the used will have to apply jnp.unpackbits(events, axis=<time axis>) prior to feeding the data to the network.
    """

    def __init__(self, encoding_dim, sample_T = 100):
        self.encoding_dim = encoding_dim
        self.sample_T = sample_T
        
    def __call__(self, events):
        # tensor has dimensions (time_steps, encoding_dim)
        tensor = np.zeros((events["t"].max()+1, self.encoding_dim), dtype=int)
        np.add.at(tensor, (events["t"], events["x"]), 1)
        #return tensor[:self.sample_T,:]
        tensor = tensor[:self.sample_T,:]
        tensor = np.minimum(tensor, 1)
        tensor = np.packbits(tensor, axis=0)
        return tensor
    

class SHD_loader():
    """
    Dataloading wrapper for the Spiking Heidelberg Dataset. The entire dataset is loaded to vRAM in a temporally compressed format. The user must
    apply jnp.unpackbits(events, axis=<time axis>) prior to feeding to an SNN. 

    https://zenkelab.org/resources/spiking-heidelberg-datasets-shd/
    """


    # Change this to allow a config dictionary of 
    def __init__(self, batch_size=256, sample_T = 100, channels=128, val_size=0.2):
        #####################################
        # Load datasets and process them using tonic.
        #####################################
        shd_timestep = 1e-6
        shd_channels = 700
        net_channels = channels
        net_dt = 1/sample_T
           
        self.batch_size = batch_size
        self.val_size = val_size
        self.obs_shape = tuple([net_channels,])
        self.act_shape = tuple([20,])
        
        transform = transforms.Compose([
        transforms.Downsample(
            time_factor=shd_timestep / net_dt,
            spatial_factor=net_channels / shd_channels
            ),
            SHD2Raster(net_channels, sample_T = sample_T)
        ])
        
        train_val_dataset = datasets.SHD("./data", train=True, transform=transform)
        test_dataset = datasets.SHD("./data", train=False, transform=transform)
        
        
        #########################################################################
        # load entire dataset to GPU as JNP Array, create methods for splits
        #########################################################################

    
        # create train/validation split here...
        # generate indices: instead of the actual data we pass in integers instead
        train_indices, val_indices = train_test_split(
            range(len(train_val_dataset)),
            test_size=self.val_size,
            random_state=0,
            shuffle=True # This really should be set externally!!!!!
        )


        train_split = Subset(train_val_dataset, train_indices)
        self.train_len = len(train_indices)

        train_dl = iter(DataLoader(train_split, batch_size=self.train_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=False))
        
        x_train, y_train = next(train_dl)
        self.x_train = jnp.array(x_train, dtype=jnp.uint8)
        self.y_train = jnp.array(y_train, dtype=jnp.uint8)
        ############################
        
        val_split = Subset(train_val_dataset, val_indices)
        self.val_len = len(val_indices)

        val_dl = iter(DataLoader(val_split, batch_size=self.val_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=False))
        
        x_val, y_val = next(val_dl)
        self.x_val = jnp.array(x_val, dtype=jnp.uint8)
        self.y_val = jnp.array(y_val, dtype=jnp.uint8)
        ##########################
        # Test set setup
        ##########################
        self.test_len = len(test_dataset)
        test_dl = iter(DataLoader(test_dataset, batch_size=self.test_len,
                          collate_fn=tonic.collation.PadTensors(batch_first=True), drop_last=True, shuffle=True))
                
        x_test, y_test = next(test_dl)
        self.x_test = jnp.array(x_test, dtype=jnp.uint8)
        self.y_test = jnp.array(y_test, dtype=jnp.uint8)


        
        @jax.jit
        def _train_epoch(shuffle_key):
            cutoff = self.train_len % self.batch_size
            
            obs = jax.random.permutation(shuffle_key, self.x_train, axis=0)[:-cutoff] # self.x_train[:-cutoff]
            labels = jax.random.permutation(shuffle_key, self.y_train, axis=0)[:-cutoff] # self.y_train[:-cutoff]
            
            obs = jnp.reshape(obs, (-1, self.batch_size) + obs.shape[1:])
            labels = jnp.reshape(labels, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
            
        self.train_epoch = _train_epoch
            
        @jax.jit
        def _val_epoch():
            cutoff = self.val_len % self.batch_size
            
            x_val = self.x_val[:-cutoff]
            y_val = self.y_val[:-cutoff]
            
            obs = jnp.reshape(x_val, (-1, self.batch_size) + x_val.shape[1:])
            labels = jnp.reshape(y_val, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
        
        self.val_epoch = _val_epoch
        
        
        @jax.jit
        def _test_epoch():
            cutoff = self.test_len % self.batch_size
            
            x_test = self.x_test[:-cutoff]
            y_test = self.y_test[:-cutoff]
            
            obs = jnp.reshape(x_test, (-1, self.batch_size) + x_test.shape[1:])
            labels = jnp.reshape(y_test, (-1, self.batch_size))
            
            return State(obs=obs, labels=labels)
        
        self.test_epoch = _test_epoch
    
    


import jax
import jax.numpy as jnp
import optax

from jax import tree_util as tree

### Change all of these to H.O.F.s

class silence_reg:
    """
    L2-Norm per-neuron activation normalization for spiking less than a target number of times.

    Attributes:
        min_spikes: neurons which spike below this value on average
        over the batch incur quadratic penalty.
    """
    def __init__(self, min_spikes):
        def _loss(x):
            return (jnp.maximum(0, min_spikes-jnp.mean(x, axis=0)))**2
        
        def _flatten(x):
            return jnp.reshape(x, (x.shape[0], -1))
        
        def _call(spikes):
            flat_spikes = tree.tree_map(_flatten, spikes)
            loss_vectors = tree.tree_map(_loss, flat_spikes)
            return jnp.sum(jnp.concatenate(tree.tree_flatten(loss_vectors)[0]))
        
        self.call = jax.jit(_call)
        
    def __call__(self, spikes):
        return self.call(spikes)
        
class sparsity_reg:
    """
    Layer activation normalization that seeks to discourage all neurons having a high firing rate.

    Attributes:
        max_spikes: Threshold for which penalty is incurred if the average number of
        spikes in the layer exceeds it.

        norm: an Optax loss function. Default is Huber loss.
    """
    def __init__(self, max_spikes, norm=optax.huber_loss):
        def _loss(x):
            return norm(jnp.maximum(0, jnp.mean(x, axis=-1) - max_spikes)) # this may not work for convolution layers....
        
        def _flatten(x):
            return jnp.reshape(x, (x.shape[0], -1))
        
        def _call(spikes):
            flat_spikes = tree.tree_map(_flatten, spikes)
            loss_vectors = tree.tree_map(_loss, flat_spikes)
            return jnp.sum(jnp.concatenate(tree.tree_flatten(loss_vectors)[0]))
        
        self.call = jax.jit(_call)
        
    def __call__(self, spikes):
        return self.call(spikes) 


@jax.jit
def integral_accuracy(traces, targets):
    """
    Calculate the accuracy of a network's predictions based on the voltage traces.
    Used in combination with a Leaky-Integrate neuron model as the final layer.

    """

    preds = jnp.argmax(jnp.sum(traces, axis=-2), axis=-1)
    return jnp.sum(preds == targets) / traces.shape[0], preds

# smoothing can be critical to the performance of your model...
# change this to be a higher-order function yielding a func with a set smoothing rate.
@jax.jit
def integral_crossentropy(traces, targets, smoothing=0.3):
    """
    Calculate the crossentropy between the integral of membrane potentials.
    Allows for label smoothing to discourage silencing 
    the other neurons in the readout layer.

    Attributes:
        traces: the output of the final layer of the SNN
        targets: the integer labels for each class
        smoothing: [optional] rate at which to smooth labels.
    """

    logits = jnp.sum(traces, axis=-2) # time axis.
    labels = optax.smooth_labels(jax.nn.one_hot(targets, logits.shape[-1]), smoothing)
    return optax.softmax_cross_entropy(logits, labels).mean() 

# convert to function that returns compiled function
def mse_spikerate(traces, targets, sparsity=0.25, smoothing=0.0):
    """
    Calculate the mean squared error of the mean spike rate.
    Allows for label smoothing to discourage silencing 
    the other neurons in the readout layer.

    Attributes:
        traces: the output of the final layer of the SNN
        targets: the integer labels for each class
        smoothing: [optional] rate at which to smooth labels.
    """
    t = traces.shape[1]
    logits = jnp.mean(traces, axis=-2) # time axis.
    labels = optax.smooth_labels(jax.nn.one_hot(targets, logits.shape[-1]), smoothing)
    return jnp.mean(optax.squared_error(logits, labels * sparsity * t))



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
long_description = "Spyx is a compact library built on top of DeepMind's Haiku library, enabling easy construction of spiking neural network models."


requires = (
    [
        "dm-haiku",
        "optax",
        "jax_tqdm",
    ],
)

# This call to setup() does all the work
setup(
    name="spyx",
    version="0.1.7",
    description="Spyx: SNNs in JAX",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kmheckel/spyx",
    author="Kade Heckel",
    author_email="example@email.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    packages=["spyx"],
    include_package_data=True,
    install_requires=requires
)



import jax
import jax.numpy as jnp
import haiku as hk
from .axn import Axon

# need to add shape checking/warning
def PopulationCode(num_classes):
    """
    Add population coding to the preceding neuron layer. Preceding layer's output shape must be a multiple of
    the number of classes. Use this for rate coded SNNs where the time steps are too few to get a good spike count.
    """
    def _pop_code(x):
        return jnp.sum(jnp.reshape(x, (-1,num_classes)), axis=-1)
    return jax.jit(_pop_code)

class ALIF(hk.RNNCore): 
    """
    Adaptive LIF Neuron based on the model used in LSNNs:

    Bellec, G., Salaj, D., Subramoney, A., Legenstein, R. & Maass, W. 
    Long short- term memory and learning-to-learn in networks of spiking neurons. 
    32nd Conference on Neural Information Processing Systems (2018).
    
    """


    def __init__(self, hidden_shape, beta=None, gamma=None, threshold=1,
                 activation = Axon(),
                 name="ALIF"):
        super().__init__(name=name)
        self.hidden_shape = hidden_shape
        self.beta = beta
        self.gamma = gamma
        self.init_threshold = threshold
        self.act = activation
    
    def __call__(self, x, VT):
        # this probably needs changed to be spltting an array
        V, T = jnp.split(VT, 2, -1)
        
        gamma = self.gamma
        beta = self.beta
        # threshold adaptation
        if not gamma:
            gamma = hk.get_parameter("gamma", self.hidden_shape, 
                                 init=hk.initializers.TruncatedNormal(0.25, 0.5))
            gamma = jax.nn.hard_sigmoid(gamma)
        if not beta:
            beta = hk.get_parameter("beta", self.hidden_shape, 
                                init=hk.initializers.TruncatedNormal(0.25, 0.5))
            beta = jax.nn.hard_sigmoid(beta)
        # calculate whether spike is generated, and update membrane potential
        thresh = self.init_threshold + T
        spikes = self.act(V - thresh)
        V = (beta*V + x - spikes*thresh).astype(jnp.float16)
        T = gamma*T + (1-gamma)*spikes
        
        VT = jnp.concatenate([V,T], axis=-1, dtype=jnp.float16)
        return spikes, VT
    
    # not sure if this is borked.
    def initial_state(self, batch_size):
        return jnp.zeros((batch_size,) + tuple(2*s for s in self.hidden_shape), dtype=jnp.float16)
         
class LI(hk.RNNCore):
    """
    Leaky-Integrate (Non-spiking) neuron model.

    Attributes:
        layer_size: Number of output neurons from the previous linear layer.

        beta: Decay rate on membrane potential (voltage). Set uniformly across the layer.
    """

    def __init__(self, layer_shape, beta=0.8, name="LI"):
        super().__init__(name=name)
        self.layer_shape = layer_shape
        self.beta = beta
    
    def __call__(self, x, Vin):
        # calculate whether spike is generated, and update membrane potential
        Vout = self.beta*Vin + x
        return Vout, Vout
    
    def initial_state(self, batch_size):
        return jnp.zeros((batch_size,) + self.layer_shape, dtype=jnp.float32)

class IF(hk.RNNCore): # bfloat16 covers a wide range of unused values...
    """
    Integrate and Fire neuron model 
    

    Attributes:
        hidden_size: Size of preceding layer's outputs
        threshold: threshold for reset. Defaults to 1.
        activation: spyx.activation function, default is Heaviside with Straight-Through-Estimation.
    """

    def __init__(self, hidden_shape, threshold=1, 
                 activation = Axon(),
                 name="LIF"):
        super().__init__(name=name)
        self.hidden_shape = hidden_shape
        self.threshold = threshold
        self.act = activation
    
    def __call__(self, x, V):
        # calculate whether spike is generated, and update membrane potential
        spikes = self.act(V - self.threshold)
        V = (V + x - spikes*self.threshold).astype(jnp.float16)
        
        return spikes, V

    def initial_state(self, batch_size): # figure out how to make dynamic...
        return jnp.zeros((batch_size,) + self.hidden_shape, dtype=jnp.float16)


class LIF(hk.RNNCore): # bfloat16 covers a wide range of unused values...
    """
    Leaky Integrate and Fire neuron model inspired by the implementation in
    snnTorch:

    https://snntorch.readthedocs.io/en/latest/snn.neurons_leaky.html
    

    Attributes:
        hidden_size: Size of preceding layer's outputs
        beta: decay rate. Set to float in range (0,1] for uniform decay across layer, otherwise it will be a normal
            distribution centered on 0.5 with stddev of 0.25
        threshold: threshold for reset. Defaults to 1.
        activation: spyx.activation function, default is Heaviside with Straight-Through-Estimation.
    """

    def __init__(self, hidden_shape: tuple, beta=None, threshold=1, 
                 activation = Axon(),
                 name="LIF"):
        super().__init__(name=name)
        self.hidden_shape = hidden_shape
        self.beta = beta
        self.threshold = threshold
        self.act = activation
    
    def __call__(self, x, V):
        
        # numerical stability gremlin...
        beta = self.beta
        if not beta:
            beta = hk.get_parameter("beta", self.hidden_shape, dtype=jnp.float16,
                                init=hk.initializers.TruncatedNormal(0.25, 0.5))
            beta = jax.nn.hard_sigmoid(beta)
            
        # calculate whether spike is generated, and update membrane potential
        spikes = self.act(V - self.threshold)
        V = (beta*V + x - spikes*self.threshold).astype(jnp.float16)
        
        return spikes, V

    def initial_state(self, batch_size): # figure out how to make dynamic...
        return jnp.zeros((batch_size,) + self.hidden_shape, dtype=jnp.float16)


class RLIF(hk.RNNCore): # bfloat16 covers a wide range of unused values...
    """
    Recurrent LIF Neuron adapted from snnTorch:

    https://snntorch.readthedocs.io/en/latest/snn.neurons_rleaky.html
    """

    def __init__(self, hidden_shape, beta=None, threshold=1,
                 activation = Axon(),
                 name="RLIF"):
        super().__init__(name=name)
        self.hidden_shape = hidden_shape
        self.beta = beta
        self.threshold = threshold
        self.act = activation
    
    def __call__(self, x, V):
        # calculate whether spike is generated, and update membrane potential
        recurrent = hk.get_parameter("w", self.hidden_shape, init=hk.initializers.TruncatedNormal())
        
        beta = self.beta
        if not beta:
            beta = hk.get_parameter("beta", self.hidden_shape, 
                                init=hk.initializers.TruncatedNormal(0.25, 0.5))
            beta = jax.nn.hard_sigmoid(beta)
        
        spikes = self.act(V - self.threshold)
        V = (beta*V + x + recurrent*spikes - spikes*self.threshold).astype(jnp.float16)
        
        return spikes, V

    def initial_state(self, batch_size):
        return jnp.zeros((batch_size,) + self.hidden_shape, dtype=jnp.float16)

# Current Based (CuBa)
class SC(hk.RNNCore): 
    """
    Conductance based neuron modeling synaptic conductance.

    Adapted from snnTorch:

    https://snntorch.readthedocs.io/en/latest/snn.neurons_synaptic.html
    """

    def __init__(self, hidden_shape, alpha=None, beta=None, threshold=1, 
                 activation = Axon(),
                 name="SC"):
        super().__init__(name=name)
        self.hidden_shape = hidden_shape
        self.alpha = alpha
        self.beta = beta
        self.threshold = threshold
        self.act = activation
    
    def __call__(self, x, VI):
        V, I = jnp.split(VI, 2, -1)
        
        alpha = self.alpha
        beta = self.beta
        # threshold adaptation
        if not alpha:
            alpha = hk.get_parameter("alpha", self.hidden_shape, 
                                 init=hk.initializers.TruncatedNormal(0.25, 0.5))
            alpha = jax.nn.hard_sigmoid(alpha)
        if not beta:
            beta = hk.get_parameter("beta", self.hidden_shape, 
                                init=hk.initializers.TruncatedNormal(0.25, 0.5))
            beta =jax.nn.hard_sigmoid(beta)
        # calculate whether spike is generated, and update membrane potential
        spikes = self.act(V - self.threshold)
        I = alpha*I + x
        V = (beta*V + I - spikes*self.threshold).astype(jnp.float16) # cast may not be needed?
        
        VI = jnp.concatenate([V,I], axis=-1, dtype=jnp.float16)
        return spikes, VI
    
    # this is probably borked with the shaping now.
    def initial_state(self, batch_size):
        return jnp.zeros((batch_size,) + tuple(2*s for s in self.hidden_shape), dtype=jnp.float16)
    
